import os
import functools
import json
from json import encoder
encoder.FLOAT_REPR = lambda o: format(o, '.8f')
from glob import glob

import numpy as np

from .activity import Measurement, SIndex, StarProps

__all__ = ['glob_spectra_paths', 'stars_to_json', 'json_to_stars']


def glob_spectra_paths(data_dir, target_names):
    """
    Collect paths to spectrum FITS files.

    Parameters
    ----------
    data_dir : str or list
        Paths to the directories containing spectrum FITS files
    target_names : list
        String patterns that match the beginning of files with targets to
        collect.

    Returns
    -------
    spectra_paths : list
        List of paths to spectrum FITS files
    """
    if type(data_dir) != list:
        data_dir = [data_dir]

    all_spectra_paths = []

    for d_dir in data_dir:

        # Collect files for each target:
        spectra_paths_lists = [glob(os.path.join(d_dir,
                                                 '{0}*.wfrmcpc.fits'.format(name)))
                               for name in target_names]

        # Reduce to one list:
        spectra_paths = functools.reduce(list.__add__, spectra_paths_lists)

        all_spectra_paths.extend(spectra_paths)

    return all_spectra_paths


def combine_measurements(measurement_list):
    mean = np.mean([m.value for m in measurement_list])
    err = np.sqrt(np.sum(np.array([m.err for m in measurement_list])**2))
    return Measurement(value=mean, err=err, meta=len(measurement_list))


def floats_to_strings(d):
    dictionary = d.copy()
    for key in dictionary:
        dictionary[key] = str(dictionary[key])
    return dictionary


def stars_to_json(star_list, output_path='star_data.json'):
    """
    Save list of stellar properties to a JSON file.

    Parameters
    ----------
    star_list : list of `StarProps`
        Star properties to save to json
    output_path : str
        File path to output
    """
    stars_attrs = star_list[0].__dict__.keys()
    all_data = dict()

    for star in star_list:
        star_data = dict()

        for attr in stars_attrs:
            value = getattr(star, attr)

            if isinstance(value, Measurement):
                value = floats_to_strings(value.__dict__)
            elif isinstance(value, SIndex):
                value = value.to_dict()
            else:
                value = str(value)

            star_data[attr] = value

        all_data[star.name + '; ' + str(star.time.datetime)] = star_data

    with open(output_path, 'w') as w:
        json.dump(all_data, w, indent=4, sort_keys=True)


def json_to_stars(json_path):
    """
    Loads JSON archive into list of `StarProps` objects.

    Parameters
    ----------
    json_path : str
        Path to saved stellar properties

    Returns
    -------
    stars : list of `StarProps`
        List of stellar properties.
    """
    with open(json_path, 'r') as w:
        dictionary = json.load(w)

    stars = [StarProps.from_dict(dictionary[star]) for star in dictionary]
    return stars




# -*- coding: utf-8 -*-
# Licensed under a 3-clause BSD style license - see LICENSE.rst
#
# Astropy documentation build configuration file.
#
# This file is execfile()d with the current directory set to its containing dir.
#
# Note that not all possible configuration values are present in this file.
#
# All configuration values have a default. Some values are defined in
# the global Astropy configuration which is loaded here before anything else.
# See astropy.sphinx.conf for which values are set there.

from aesop import __version__
import sys
import datetime

try:
    from sphinx_astropy.conf.v2 import *  # noqa
except ImportError:
    print('ERROR: the documentation requires the sphinx-astropy package to '
          'be installed')
    sys.exit(1)

# -- General configuration ----------------------------------------------------
# By default, highlight as Python 3.
highlight_language = 'python3'

# If your documentation needs a minimal Sphinx version, state it here.
# needs_sphinx = '1.7'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns.append('_templates')  # noqa
# Exclude template PSF block specification documentation
exclude_patterns.append('psf_spec/*')  # noqa

plot_formats = ['png', 'hires.png', 'pdf', 'svg']


# -- Project information ------------------------------------------------------

# This does not *have* to match the package name, but typically does
project = 'aesop'
author = 'Brett Morris'
copyright = '{0}, {1}'.format(
    datetime.datetime.now().year, author)

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
__import__(project)
package = sys.modules[project]

# The full version, including alpha/beta/rc tags.
release = __version__
dev = "dev" in __version__
# The short X.Y version.
version = '.'.join(release.split('.')[:2])

# -- Options for HTML output ---------------------------------------------------

# A NOTE ON HTML THEMES
# The global astropy configuration uses a custom theme, 'bootstrap-astropy',
# which is installed along with astropy. A different theme can be used or
# the options for this theme can be modified by overriding some of the
# variables set in the global configuration. The variables set in the
# global configuration are listed below, commented out.

# Add any paths that contain custom themes here, relative to this directory.
# To use a different custom theme, add the directory containing the theme.
#html_theme_path = []

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes. To override the custom theme, set this to the
# name of a builtin theme or the name of a custom theme in html_theme_path.
# html_theme = None

# Custom sidebar templates, maps document names to template names.
#html_sidebars = {}

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
#html_favicon = ''

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
#html_last_updated_fmt = ''

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
html_title = '{0}'.format(project, version)

# Output file base name for HTML help builder.
htmlhelp_basename = project + 'doc'

# Prefixes that are ignored for sorting the Python module index
modindex_common_prefix = ["aesop."]

# -- Options for LaTeX output --------------------------------------------------

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, documentclass [howto/manual]).
latex_documents = [('index', project + '.tex', project + u' Documentation',
                    author, 'manual')]


# -- Options for manual page output --------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [('index', project.lower(), project + u' Documentation',
              [author], 1)]


# -- Options for the edit_on_github extension ----------------------------------

# Add additional Sphinx extensions:
extensions += [
    'matplotlib.sphinxext.plot_directive',
    'sphinx.ext.graphviz'
]

# -- Resolving issue number to links in changelog -----------------------------
github_issues_url = 'https://github.com/bmorris3/aesop/issues/'

release = __version__
dev = "dev" in release

html_copy_source = False

html_theme_options.update(  # noqa: F405
    {
        "github_url": "https://github.com/bmorris3/aesop",
        "external_links": [
            {"name": "astropy docs", "url": "https://docs.astropy.org/en/stable/"},
        ],
        "use_edit_page_button": True,
    }
)

html_context = {
    "default_mode": "light",
    "to_be_indexed": ["stable", "latest"],
    "is_development": dev,
    "github_user": "bmorris3",
    "github_repo": "aesop",
    "github_version": "main",
    "doc_path": "docs",
}

#
# Some warnings are impossible to suppress, and you can list specific references
# that should be ignored in a nitpick-exceptions file which should be inside
# the docs/ directory. The format of the file should be:
#
# <type> <class>
#
# for example:
#
# py:class astropy.io.votable.tree.Element
# py:class astropy.io.votable.tree.SimpleElement
# py:class astropy.io.votable.tree.SimpleElementWithContent
#
# Uncomment the following lines to enable the exceptions:
#
# for line in open('nitpick-exceptions'):
#     if line.strip() == "" or line.startswith("#"):
#         continue
#     dtype, target = line.split(None, 1)
#     target = target.strip()
#     nitpick_ignore.append((dtype, six.u(target)))




from .catalog import get_k2_epic_catalog

from astroquery.simbad import Simbad

__all__ = ['query_for_spectral_type', 'query_for_T_eff']

# Source: http://www.uni.edu/morgans/astro/course/Notes/section2/spectraltemps.html
effective_temperatures = {'A0': 9600,
 'A1': 9330,
 'A2': 9040,
 'A3': 8750,
 'A4': 8480,
 'A5': 8310,
 'A7': 7920,
 'B0': 29200,
 'B1': 23000,
 'B2': 21000,
 'B3': 17600,
 'B5': 15200,
 'B6': 14300,
 'B7': 13500,
 'B8': 12300,
 'B9': 11400,
 'F0': 7350,
 'F2': 7050,
 'F3': 6850,
 'F5': 6700,
 'F6': 6550,
 'F7': 6400,
 'F8': 6300,
 'G0': 6050,
 'G1': 5930,
 'G2': 5800,
 'G5': 5660,
 'G8': 5440,
 'K0': 5240,
 'K1': 5110,
 'K2': 4960,
 'K3': 4800,
 'K4': 4600,
 'K5': 4400,
 'K7': 4000,
 'L0': 2600,
 'L3': 2200,
 'L8': 1500,
 'M0': 3750,
 'M1': 3700,
 'M2': 3600,
 'M3': 3500,
 'M4': 3400,
 'M5': 3200,
 'M6': 3100,
 'M7': 2900,
 'M8': 2700,
 'O5': 54000,
 'O6': 45000,
 'O7': 43300,
 'O8': 40600,
 'O9': 37800,
 'T2': 1400,
 'T6': 1000,
 'T8': 800}


def query_for_spectral_type(identifier, only_first_two_characters=True,
                            default_sptype='G0'):
    """
    Search SIMBAD for the spectral type of a star.

    If no spectral type is found, the default return value is ``"G0"``.

    Parameters
    ----------
    identifier : str
        Name of target
    only_first_two_characters : bool
        Return only first two characters of spectral type?
    default_sptype : str
        Spectral type returned when none is found on SIMBAD

    Returns
    -------
    sptype : str
        Spectral type of the star.
    """
    customSimbad = Simbad()
    customSimbad.SIMBAD_URL = 'http://simbad.harvard.edu/simbad/sim-script'
    customSimbad.add_votable_fields('sptype')
    result = customSimbad.query_object(identifier)

    if len(result['SP_TYPE']) > 0:

        if only_first_two_characters:
            return result['SP_TYPE'][0][:2].strip().decode()
        else:
            return result['SP_TYPE'][0].decode()
    else:
        return default_sptype.decode()


def query_for_T_eff(identifier):
    """
    Get the approximate effective temperature of a star.

    Query SIMBAD for the spectral type of the target, convert
    spectral type to approximate effective temperature, in general.
    If the target is in the K2 EPIC, use the EPIC Teff.

    Parameters
    ----------
    identifier : str
        Name of target

    Returns
    -------
    T_eff : int
        Approximate effective temperature of the star.
    """
    if not identifier.startswith('EPIC'):
        sptype = query_for_spectral_type(identifier)
        while not sptype in effective_temperatures:
            letter, number = list(sptype)
            sptype = letter + str(int(number) - 1)

        T_eff = effective_temperatures[sptype]
    else:
        k2_epic_table = get_k2_epic_catalog()
        epic_number = int(identifier[4:]) # Remove the EPIC, make int
        T_eff = k2_epic_table.loc[epic_number]['Teff']

    return T_eff





import numpy as np
import astropy.units as u
import pytest
from astropy.units import UnitsError
from astropy.utils.data import download_file

from ..spectra import Spectrum1D, EchelleSpectrum


def test_constructor():
    wl = np.linspace(3000, 4000) * u.Angstrom
    f = 1 + 0.1 * np.random.randn()
    example = Spectrum1D(wavelength=wl, flux=f)

    # raise an error if there is no unit on the wavelength
    with pytest.raises(TypeError):
        example = Spectrum1D(wavelength=[0, 1, 2, 3], flux=f)

    # raise an error if the wavelength unit is not a unit of length
    with pytest.raises(UnitsError):
        example = Spectrum1D(wavelength=u.Quantity([0, 1, 2, 3], unit=u.day),
                             flux=f)


@pytest.mark.remote_data
def test_read_fits():
    url = ('https://stsci.box.com/shared/static/'
           'mu4fa1fmq1lw8boem12e2umyi99skbdl.fits')

    path = download_file(url, show_progress=False)

    echelle_spectrum = EchelleSpectrum.from_fits(path, format='iraf')

    assert hasattr(echelle_spectrum, 'header')
    assert hasattr(echelle_spectrum, 'time')
    assert hasattr(echelle_spectrum, 'name')
    assert str(echelle_spectrum) == ('<EchelleSpectrum: 107 orders, '
                                     '3506.8-10612.4 Angstrom>')

    # There should be more flux in the 40th order (redder) than 0th (bluer)
    assert echelle_spectrum[40].flux.mean() > echelle_spectrum[0].flux.mean()


def generate_target_standard_pairs():
    # realistic wavelength range, resolution for one order
    wl = np.linspace(6000, 6117, 1650) * u.Angstrom
    sig = 25 + 10 * (0.5 - np.random.rand())
    amp = 500 + 200 * (0.5 - np.random.rand())
    flux = amp * np.exp(-0.5 * (wl.value - wl.value.mean())**2 / sig**2)
    flux += 10 * np.random.randn(flux.shape[0])

    standard = Spectrum1D(wavelength=wl, flux=flux)

    for i in range(4):
        flux -= 100 * np.exp(-0.5 * (wl.value - wl.value.mean() -
                                     20 * (i - 1.5))**2 / 0.7**2)

    target = Spectrum1D(wavelength=wl, flux=flux)

    return target, standard


def test_continuum_norm():
    target_orders = []
    standard_orders = []
    for i in range(10):
        target, standard = generate_target_standard_pairs()
        target_orders.append(target)
        standard_orders.append(standard)

    target_spectrum = EchelleSpectrum(target_orders)
    standard_spectrum = EchelleSpectrum(standard_orders)

    # Make sure that the continuum normalization is always good to 2%
    polynomial_order = 8
    target_spectrum.continuum_normalize_from_standard(standard_spectrum,
                                                      polynomial_order)

    for order in target_spectrum.spectrum_list:
        assert abs(np.median(order.flux) - 1) < 0.02

    # Make sure spectral features are preserved in target spectra
    for order in target_spectrum.spectrum_list:
        assert np.mean(order.flux) < 1


# This file is used to configure the behavior of pytest when using the Astropy
# test infrastructure. It needs to live inside the package in order for it to
# get picked up when running the tests inside an interpreter using
# packagename.test

try:
    from pytest_astropy_header.display import (PYTEST_HEADER_MODULES,
                                               TESTED_VERSIONS)
    ASTROPY_HEADER = True
except ImportError:
    ASTROPY_HEADER = False


def pytest_configure(config):
    if ASTROPY_HEADER:
        config.option.astropy_header = True

        # Customize the following lines to add/remove entries from the
        # list of packages for which version numbers are displayed when
        # running the tests.
        PYTEST_HEADER_MODULES['Numpy'] = 'numpy'
        PYTEST_HEADER_MODULES['Astropy'] = 'astropy'
        PYTEST_HEADER_MODULES.pop('h5py', None)



"""
Tools for organizing, normalizing echelle spectra.
"""

import matplotlib.pyplot as plt
import numpy as np
from scipy.ndimage import gaussian_filter1d
from scipy.optimize import least_squares
from scipy.stats import binned_statistic

from astropy.io import fits
import astropy.units as u
import astropy.constants as c
from astropy.time import Time
from astropy.stats import mad_std

from astropy.coordinates.representation import (CartesianRepresentation,
                                                UnitSphericalRepresentation)
from astropy.coordinates import SkyCoord, solar_system, EarthLocation

from specutils import SpectrumCollection
from .spectral_type import query_for_T_eff
from .phoenix import get_phoenix_model_spectrum
from .masking import get_spectrum_mask
from .activity import true_h_centroid, true_k_centroid

__all__ = ["EchelleSpectrum", "slice_spectrum", "interpolate_spectrum",
           "cross_corr", "Spectrum1D"]


class Spectrum1D(object):
    """
    Simple 1D spectrum object.

    A ``Spectrum1D`` object can be used to describe one order of an echelle
    spectrum, for example.

    If the spectrum is initialized with ``wavelength``s that are not strictly
    increasing, ``Spectrum1D`` will sort the ``wavelength``, ``flux`` and
    ``mask`` arrays so that ``wavelength`` is monotonically increasing.
    """
    @u.quantity_input(wavelength=u.Angstrom)
    def __init__(self, wavelength=None, flux=None, name=None, mask=None,
                 wcs=None, meta=dict(), time=None, continuum_normalized=None):
        """
        Parameters
        ----------
        wavelength : `~numpy.ndarray`
            Wavelengths
        flux : `~numpy.ndarray`
            Fluxes
        name : str (optional)
            Name for the spectrum
        mask : `~numpy.ndarray` (optional)
            Boolean mask of the same shape as ``flux``
        wcs : `~specutils.Spectrum1DLookupWCS` (optional)
            Store the WCS parameters
        meta : dict (optional)
            Metadata dictionary.
        continuum_normalized : bool (optional)
            Is this spectrum continuum normalized?
        """

        # Are wavelengths stored in increasing order?
        wl_inc = np.all(np.diff(wavelength) > 0)

        # If not, force them to be, to simplify linear interpolation later.
        if not wl_inc:
            wl_sort = np.argsort(wavelength)
            wavelength = wavelength[wl_sort]
            flux = flux[wl_sort]
            if mask is not None:
                mask = mask[wl_sort]

        self.wavelength = wavelength
        self.wavelength_unit = wavelength.unit
        self.flux = flux if hasattr(flux, 'unit') else u.Quantity(flux)
        self.name = name
        self.mask = mask
        self.wcs = wcs
        self.meta = meta
        self.time = time
        self.continuum_normalized = continuum_normalized

    def flux_calibrate_parameters(self, flux_calibrated_spectrum, polynomial_order, plots=False):
        """
        Interpolate high-res spectrum to low-res flux calibrated spectrum, then fit
        the ratio with a polynomial to flux calibrate. Returns polynomial coefficients

        Parameters
        ----------
        flux_calibrated_spectrum : `~aesop.Spectrum1D`
            Already flux calibrated low-resolution spectrum of the same object
        polynomial_order : int
            Order of polynomial fit
        plots : bool
            If True, plot the sensitivity data and the fit

        Returns
        -------
        fit_params : `~numpy.ndarray`
            Best-fit polynomial coefficients
        """

        int_spectrum = interpolate_spectrum(spectrum=self, new_wavelengths=flux_calibrated_spectrum.wavelength)

        sens_data = flux_calibrated_spectrum.flux/int_spectrum.flux

        fit_params = np.polyfit(int_spectrum.wavelength, sens_data, polynomial_order)

        if plots:
            plt.figure()
            plt.plot(int_spectrum.wavelength,
                     sens_data,label='Data')
            plt.plot(int_spectrum.wavelength,
                     np.polyval(fit_params,
                                int_spectrum.wavelength),label='Fit')
            plt.gca().set(xlabel='Wavelength [{0}]'.format(self.wavelength_unit),
               ylabel='1/Sensitivity')
            plt.legend()
            plt.show()

        return fit_params

    def flux_calibrate(self, flux_calibrated_spectrum, polynomial_order):
        """
        Calculates coefficients of sensitivity function, then returns flux-calibrated spectrum

        Parameters
        ----------
        flux_calibrated_spectrum : `~aesop.Spectrum1D`
            Already flux calibrated low-resolution spectrum of the same object
        polynomial_order : int
            Order of polynomial fit

        Returns
        -------
        transformed_spectrum : `~aesop.Spectrum1D`
            Spectrum transformed with sensitivity polynomial
        """

        sens_params = self.flux_calibrate_parameters(flux_calibrated_spectrum, polynomial_order)

        sens = np.polyval(sens_params,self.wavelength)

        calibrated_flux = self.flux * sens

        transformed_spectrum = Spectrum1D(wavelength=self.wavelength, flux=calibrated_flux)

        return transformed_spectrum

    def plot(self, ax=None, **kwargs):
        """
        Plot the spectrum.

        Parameters
        ----------
        ax : `~matplotlib.axes.Axes` (optional)
            The `~matplotlib.axes.Axes` to draw on, if provided.
        kwargs
            All other keyword arguments are passed to `~matplotlib.pyplot.plot`

        """
        if ax is None:
            ax = plt.gca()

        ax.plot(self.masked_wavelength, self.masked_flux, **kwargs)
        ax.set(xlabel='Wavelength [{0}]'.format(self.wavelength_unit),
               ylabel='Flux')
        if self.name is not None:
            ax.set_title(self.name)

    @property
    def masked_wavelength(self):
        if self.mask is not None:
            return self.wavelength[np.logical_not(self.mask)]
        else:
            return self.wavelength

    @property
    def masked_flux(self):
        if self.mask is not None:
            return self.flux[np.logical_not(self.mask)]
        else:
            return self.flux

    @classmethod
    def from_specutils(cls, spectrum1d, name=None, **kwargs):
        """
        Convert a `~specutils.Spectrum1D` object into our Spectrum1D object.

        Parameters
        ----------
        spectrum1d : `~specutils.Spectrum1D`
            Input spectrum
        name : str
            Target/spectrum name
        """
        return cls(wavelength=spectrum1d.wavelength, flux=spectrum1d.flux,
                   mask=spectrum1d._mask, name=name, **kwargs)

    @classmethod
    def from_array(cls, wavelength, flux, dispersion_unit=None, name=None,
                   **kwargs):
        """
        Initialize a spectrum with the same call signature as
        `~specutils.Spectrum1D.from_array`.

        Parameters
        ----------
        wavelength : `~astropy.units.Quantity`
            Spectrum wavelengths
        flux : `~astropy.units.Quantity` or `~numpy.ndarray`
            Spectrum fluxes
        dispersion_unit : `~astropy.units.Unit` (optional)
            Unit of the wavelength
        name : str (optional)
            Name of the target/spectrum
        """
        if not hasattr(wavelength, 'unit') and dispersion_unit is not None:
            wavelength = wavelength * dispersion_unit
        return cls(wavelength=wavelength, flux=flux, name=name, **kwargs)

    def __repr__(self):
        wl_unit = u.Angstrom
        min_wavelength = self.wavelength.min()
        max_wavelength = self.wavelength.max()

        if self.name is not None:
            name_str = '"{0}" '.format(self.name)
        else:
            name_str = ''

        return ("<Spectrum1D: {0}{1:.1f}-{2:.1f} {3}>"
                .format(name_str, min_wavelength.to(wl_unit).value,
                        max_wavelength.to(wl_unit).value, wl_unit))

    def mask_outliers(self, reject_negative=True, mad_clip=True,
                      mad_outlier_factor=3):
        """
        Identify outliers, update the ``mask`` attribute.

        Parameters
        ----------
        reject_negative : bool (optional)
            Reject fluxes < -0.5. Default is `True`.
        mad_clip : bool
            Reject fluxes more than ``mad_outlier_factor`` times the median
            absolute deviation (MAD) from the continuum flux.
        mad_outlier_factor : float
            MAD-masking factor -- fluxes more than ``mad_outlier_factor`` away
            from the continuum flux will be masked.
        """
        outliers = np.zeros_like(self.flux.value).astype(bool)

        if mad_clip:
            # Compute binned mean flux for outlier masking
            bs = binned_statistic(self.wavelength.value, self.flux.value,
                                  bins=300, statistic='median')
            bincenters = 0.5 * (bs.bin_edges[1:] + bs.bin_edges[:-1])
            binmedians = bs.statistic
            median_interp = np.interp(self.wavelength.value,
                                      bincenters, binmedians)

            mad = mad_std(abs(median_interp - self.flux.value))

            outliers |= (self.flux.value > mad_outlier_factor * mad +
                         np.median(self.flux.value))

        if reject_negative:
            # Also mask outliers that are very low flux
            outliers |= self.flux.value < -0.5

        self.mask |= outliers



class EchelleSpectrum(object):
    """
    Echelle spectrum of one or more spectral orders.

    The spectral orders will be indexed in order of increasing wavelength.
    """
    def __init__(self, spectrum_list, header=None, name=None, fits_path=None,
                 time=None):
        """
        Parameters
        ----------
        spectrum_list : list of `~aesop.Spectrum1D` objects
            List of `~aesop.Spectrum1D` objects for the spectra in each echelle
            order.
        header : `astropy.io.fits.header.Header` (optional)
            FITS header object associated with the echelle spectrum.
        name : str (optional)
            Name of the target or a name for the spectrum
        fits_path : str  (optional)
            Path where FITS file was opened from.
        time : `~astropy.time.Time` (optional)
            Time at which the spectrum was taken
        """
        # Sort the spectra in the list in order of increasing wavelength
        self.spectrum_list = sorted(spectrum_list,
                                    key=lambda x: x.wavelength.min())
        self.header = header
        self.name = name
        self.fits_path = fits_path
        self.standard_star_props = {}
        self.model_spectrum = None

        if header is not None and time is None:
            if 'JD' in header:
                time = Time(header['JD'], format='jd')
            elif 'DATE-OBS' in header:
                time = Time(header['DATE-OBS'], format='isot', scale='tai')

        self.time = time

    @classmethod
    def from_fits(cls, path, format=None):
        """
        Load an echelle spectrum from a FITS file.

        Parameters
        ----------
        path : str
            Path to the FITS file
        """
        spectrum_list = [Spectrum1D.from_specutils(s)
                         for s in SpectrumCollection.read(path, format=format)]
        header = fits.getheader(path)

        name = header.get('OBJNAME', None)
        return cls(spectrum_list, header=header, name=name, fits_path=path)

    def get_order(self, order):
        """
        Get the spectrum from a specific spectral order

        Parameters
        ----------
        order : int
            Echelle order to return

        Returns
        -------
        spectrum : `~specutils.Spectrum1D`
            One order from the echelle spectrum
        """
        return self.spectrum_list[order]

    def __getitem__(self, index):
        return self.spectrum_list[index]

    def __len__(self):
        return len(self.spectrum_list)

    def fit_order(self, spectral_order, polynomial_order, plots=False):
        """
        Fit a spectral order with a polynomial.

        Ignore fluxes near the CaII H & K wavelengths.

        Parameters
        ----------
        spectral_order : int
            Spectral order index
        polynomial_order : int
            Polynomial order

        Returns
        -------
        fit_params : `~numpy.ndarray`
            Best-fit polynomial coefficients
        """
        spectrum = self.get_order(spectral_order)
        mean_wavelength = spectrum.wavelength.mean()

        mask_wavelengths = ((abs(spectrum.wavelength - true_h_centroid) > 6.5*u.Angstrom) &
                            (abs(spectrum.wavelength - true_k_centroid) > 6.5*u.Angstrom))

        fit_params = np.polyfit(
            (spectrum.wavelength[mask_wavelengths] - mean_wavelength).value,
            spectrum.flux[mask_wavelengths].value, polynomial_order)

        if plots:
            plt.figure()
            # plt.plot(spectrum.wavelength, spectrum.flux)
            plt.plot(spectrum.wavelength[mask_wavelengths],
                     spectrum.flux[mask_wavelengths])
            plt.plot(spectrum.wavelength,
                     np.polyval(fit_params,
                                (spectrum.wavelength - mean_wavelength).value))
            plt.xlabel('Wavelength [{0}]'.format(spectrum.wavelength_unit))
            plt.ylabel('Flux')
            plt.show()
        return fit_params
    
    def predict_continuum(self, spectral_order, fit_params):
        """
        Predict continuum spectrum given results from a polynomial fit from
        `EchelleSpectrum.fit_order`.

        Parameters
        ----------
        spectral_order : int
            Spectral order index
        fit_params : `~numpy.ndarray`
            Best-fit polynomial coefficients

        Returns
        -------
        flux_fit : `~numpy.ndarray`
            Predicted flux in the continuum for this order
        """
        spectrum = self.get_order(spectral_order)
        mean_wavelength = spectrum.wavelength.mean()
        flux_fit = np.polyval(fit_params,
                              (spectrum.wavelength - mean_wavelength).value)
        return flux_fit

    def continuum_normalize_from_standard(self, standard_spectrum,
                                          polynomial_order, only_orders=None,
                                          plot_masking=False, plot_fit=False):
        """
        Normalize the spectrum by a polynomial fit to the standard's
        spectrum.

        Parameters
        ----------
        standard_spectrum : `EchelleSpectrum`
            Spectrum of the standard object
        polynomial_order : int
            Fit the standard's spectrum with a polynomial of this order
        only_orders : `~numpy.ndarray`
            Only do the continuum normalization for these echelle orders.
        plot_masking : bool
            Plot the masked-out low S/N regions
        plot_fit : bool
            Plot the polynomial fit to the standard star spectrum
        """

        # Copy some attributes of the standard star's EchelleSpectrum object into
        # a dictionary on the target star's EchelleSpectrum object.
        attrs = ['name', 'fits_path', 'header']
        for attr in attrs:
            self.standard_star_props[attr] = getattr(standard_spectrum, attr)

        if only_orders is None:
            only_orders = range(len(self.spectrum_list))

        for spectral_order in only_orders:
            # Extract one spectral order at a time to normalize
            standard_order = standard_spectrum.get_order(spectral_order)
            target_order = self.get_order(spectral_order)

            target_mask = get_spectrum_mask(standard_order, plot=plot_masking)

            # Fit the standard's flux in this order with a polynomial
            # fit_params = standard_spectrum.fit_order(spectral_order,
            #                                          polynomial_order)

            fit_params = standard_spectrum.fit_order(spectral_order,
                                                     polynomial_order,
                                                     plots=plot_fit)

            # Normalize the target's flux with the continuum fit from the standard
            target_continuum_fit = self.predict_continuum(spectral_order,
                                                          fit_params)

            target_continuum_normalized_flux = target_order.flux / target_continuum_fit

            normalized_target_spectrum = Spectrum1D(wavelength=target_order.wavelength,
                                                    flux=target_continuum_normalized_flux,
                                                    wcs=target_order.wcs,
                                                    mask=target_mask,
                                                    continuum_normalized=True)
            normalized_target_spectrum.meta['normalization'] = target_continuum_fit

            # Replace this order's spectrum with the continuum-normalized one
            self.spectrum_list[spectral_order] = normalized_target_spectrum

    def continuum_normalize_lstsq(self, polynomial_order, only_orders=None,
                                  plot=False, fscale_mad_factor=0.2):
        """
        Normalize the spectrum with a robust least-squares polynomial fit to the
        spectrum of each order.

        Parameters
        ----------
        polynomial_order : int
            Fit the standard's spectrum with a polynomial of this order
        only_orders : `~numpy.ndarray` (optional)
            Only do the continuum normalization for these echelle orders.
        plot_masking : bool (optional)
            Plot the masked-out low S/N regions
        plot_fit : bool (optional)
            Plot the polynomial fit to the standard star spectrum
        fscale_mad_factor : float (optional)
            The robust least-squares fitter will reject outliers by keeping
            the standard deviation of inliers close to ``fscale_mad_factor``
            times the median absolute deviation (MAD) of the fluxes.
        """
        if only_orders is None:
            only_orders = range(len(self.spectrum_list))

        for spectral_order in only_orders:
            # Extract one spectral order at a time to normalize
            s = self.get_order(spectral_order)

            x0 = np.concatenate([np.zeros(polynomial_order),
                                 [s.flux.value.mean()]])
            fscale = fscale_mad_factor * mad_std(s.flux.value)
            args = (s.wavelength.value, s.flux.value)
            res_lsq = least_squares(_residuals, x0, args=args)

            model_simple = _poly_model(res_lsq.x, args[0])

            res_robust = least_squares(_residuals, x0, loss='cauchy',
                                       f_scale=fscale, args=args)

            model_robust = _poly_model(res_robust.x, args[0])

            target_continuum_normalized_flux = s.flux / model_robust

            normalized_target_spectrum = Spectrum1D(wavelength=s.wavelength,
                                                    flux=target_continuum_normalized_flux,
                                                    wcs=s.wcs, mask=s.mask,
                                                    continuum_normalized=True)

            # Replace this order's spectrum with the continuum-normalized one
            self.spectrum_list[spectral_order] = normalized_target_spectrum

            if plot:
                fig, ax = plt.subplots(1, 2, figsize=(10, 4))

                ax[0].set_title('standard star only')
                ax[0].plot(s.wavelength.value, s.flux.value, color='k')
                ax[0].plot(s.wavelength.value, model_simple, color='DodgerBlue',
                           lw=3, label='simple lstsq')
                ax[0].plot(s.wavelength.value, model_robust, color='r', lw=3,
                           label='robust lstsq')
                ax[0].legend()

                ax[1].set_title('continuum normalized (robust polynomial)')
                ax[1].plot(s.wavelength, s.flux.value/model_robust, color='k')

    def offset_wavelength_solution(self, wavelength_offset):
        """
        Offset the wavelengths by a constant amount in each order.

        Parameters
        ----------
        wavelength_offset : `~astropy.units.Quantity` or list
            Offset the wavelengths by this amount. If ``wavelength_offset`` is a
            list, each value will be treated as an offset for on echelle order,
            otherwise a single ``wavelength_offset`` will be applied to every
            order.
        """
        if hasattr(wavelength_offset, '__len__'):
            for spectrum, offset in zip(self.spectrum_list, wavelength_offset):
                spectrum.wavelength += offset
        else:
            # Old behavior
            for spectrum in self.spectrum_list:
                spectrum.wavelength += wavelength_offset

    def rv_wavelength_shift(self, spectral_order, T_eff=None, plot=False):
        """
        Solve for the radial velocity wavelength shift.

        Parameters
        ----------
        spectral_order : int
            Echelle spectrum order to shift
        """
        order = self.spectrum_list[spectral_order]

        if self.model_spectrum is None:
            if T_eff is None:
                T_eff = query_for_T_eff(self.name)
            self.model_spectrum = get_phoenix_model_spectrum(T_eff)

        model_slice = slice_spectrum(self.model_spectrum,
                                     order.masked_wavelength.min(),
                                     order.masked_wavelength.max(),
                                     norm=order.masked_flux.max())

        delta_lambda_obs = np.abs(np.diff(order.wavelength.value[0:2]))[0]
        delta_lambda_model = np.abs(np.diff(model_slice.wavelength.value[0:2]))[0]
        smoothing_kernel_width = delta_lambda_obs/delta_lambda_model

        interp_target_slice = interpolate_spectrum(order,
                                                   model_slice.wavelength)

        rv_shift = cross_corr(interp_target_slice, model_slice,
                              kernel_width=smoothing_kernel_width)

        if plot:
            plt.figure()
            plt.plot(order.masked_wavelength + rv_shift, order.masked_flux,
                     label='shifted spectrum')
            # plt.plot(interp_target_slice.wavelength, interp_target_slice.flux,
            #          label='spec interp')
            # plt.plot(model_slice.wavelength,
            #          gaussian_filter1d(model_slice.flux, smoothing_kernel_width),
            #          label='smoothed model')
            plt.plot(model_slice.wavelength,
                     gaussian_filter1d(model_slice.flux, smoothing_kernel_width),
                     label='smooth model')

            plt.legend()
            plt.show()

        return rv_shift
    
    def barycentric_correction(self, time=None, skycoord=None, location=None):
        
        """
        Barycentric velocity correction, code from StuartLittlefair
        (https://gist.github.com/StuartLittlefair/5aaf476c5d7b52d20aa9544cfaa936a1)
        
        Uses the ephemeris set with  ``astropy.coordinates.solar_system_ephemeris.set`` 
        for corrections.
        
        For more information see `~astropy.coordinates.solar_system_ephemeris`.
        
        Will attempt to get the necessary info from the header if possible, otherwise requires time, 
        skycoord, and location parameters to be set.
  
        Parameters
        ----------
        time : `~astropy.time.Time`
            The time of observation, optional
        skycoord: `~astropy.coordinates.SkyCoord`
            The sky location to calculate the correction for, optional.
        location: `~astropy.coordinates.EarthLocation`, optional
            The location of the observatory to calculate the correction for.
    
        Returns
        -------
            barycentric_velocity : `~astropy.units.Quantity`
                The velocity correction that was added to the wavelength arrays of each order.
        """
        
        if self.time is not None:
            
            time = self.time
            
        else:
            
            assert time is not None, "Please provide a time."
        
        if self.header is not None:
            
            header = self.header
            
            if ('RA' in header) & ('DEC' in header) & ('EQUINOX' in header):
                
                if 'RADECSYS' in header: #assumes ICRS if not specified
                    frame=header['RADECSYS'].lower()
                else:
                    frame='icrs'
                
                skycoord = SkyCoord(header['RA'], header['DEC'], unit=(u.hourangle, u.deg), frame=frame,  equinox=Time(header['EQUINOX'],format='jyear'))
                
            elif skycoord is None:
                
                raise KeyError("Either set 'RA', 'DEC','RADECSYS', 'EQUINOX' header keywords or provide a location")
            
            if 'OBSERVAT' in header:
                
                location = EarthLocation.of_site(header['OBSERVAT'])
                
            elif 'SITENAME' in header:
                
                location = EarthLocation.of_site(header['SITENAME'])
            
            elif location is None:
                
                raise KeyError("Either set 'OBSERVAT' header keyword or provide a location")
                
        else:
            
            assert (skycoord is not None) & (location is not None), "You need to manually provide object coordinates and observatory location."
            
        
        ep, ev = solar_system.get_body_barycentric_posvel('earth', time) # ICRS position and velocity of Earth's geocenter
        
        op, ov = location.get_gcrs_posvel(time) # GCRS position and velocity of observatory
        
        velocity = ev + ov # ICRS and GCRS are axes-aligned. Can add the velocities.
        
        sc_cartesian = skycoord.icrs.represent_as(UnitSphericalRepresentation).represent_as(CartesianRepresentation) #Put skycoord in same frame as velocity so we can get velocity component towards object
        
        barycentric_velocity = sc_cartesian.dot(velocity).to(u.km/u.s)
        #Velocity of earth, to be added directly to wavelength. So + should result in a redshift
        redshift = barycentric_velocity/c.c 
        
        for spectrum in self.spectrum_list:
            spectrum.wavelength *= (1.0 + redshift)
            
        return barycentric_velocity
        

    def rv_wavelength_shift_ransac(self, min_order=10, max_order=45,
                                   T_eff=4700):
        """
        Solve for the radial velocity wavelength shift of every order in the
        echelle spectrum, then do a RANSAC (outlier rejecting) linear fit to the
        wavelength correction between orders ``min_order`` and ``max_order``.

        Parameters
        ----------
        min_order : int
            Index of the bluest order to fit in the wavelength correction
        max_order : int
            Index of the reddest order to fit in the wavelength correction
        T_eff : int
            Effective temperature of the PHOENIX model atmosphere to use in
            the cross-correlation.

        Returns
        -------
        wl : `~astropy.units.Quantity`
            Wavelength corrections for each order.
        """
        from sklearn import linear_model

        rv_shifts = u.Quantity([self.rv_wavelength_shift(order, T_eff=T_eff)
                                for order in range(len(self.spectrum_list))])
        X = np.arange(len(rv_shifts))[min_order:max_order, np.newaxis]
        y = rv_shifts.value[min_order:max_order]

        ransac = linear_model.RANSACRegressor()
        ransac.fit(X, y)
        line_y_ransac = ransac.predict(np.arange(len(rv_shifts))[:, np.newaxis])
        return line_y_ransac*u.Angstrom

    def __repr__(self):
        wl_unit = u.Angstrom
        min_wavelength = min([s.wavelength.min() for s in self.spectrum_list])
        max_wavelength = max([s.wavelength.max() for s in self.spectrum_list])
        return ("<EchelleSpectrum: {0} orders, {1:.1f}-{2:.1f} {3}>"
                .format(len(self.spectrum_list),
                        min_wavelength.to(wl_unit).value,
                        max_wavelength.to(wl_unit).value, wl_unit))

    def to_Spectrum1D(self):
        """
        Convert this echelle spectrum into a simple 1D spectrum.

        In wavelength regions where two spectral orders overlap, take the mean
        of the overlapping region.

        Parameters
        ----------
        mad_outlier_factor : float
            Mask positive outliers ``max_outlier_factor`` times the median
            absolute deviation plus the median of the fluxes.

        Returns
        -------
        spectrum : `~aesop.Spectrum1D`
            Simple 1D spectrum.
        """
        nonoverlapping_wavelengths = []
        nonoverlapping_fluxes = []
        dispersion_unit = None

        for i in range(1, len(self.spectrum_list) - 1):
            current_order = self.spectrum_list[i]

            previous_order = self.spectrum_list[i-1]

            next_order = self.spectrum_list[i+1]

            previous_max = previous_order.masked_wavelength.max()
            current_min = current_order.masked_wavelength.min()
            current_max = current_order.masked_wavelength.max()
            next_min = next_order.masked_wavelength.min()

            # Find the non-overlapping parts of each order, and add them to the
            # non-overlapping list
            nonoverlapping = ((current_order.masked_wavelength > previous_max) &
                              (current_order.masked_wavelength < next_min))

            if dispersion_unit is None:
                dispersion_unit = current_order.masked_wavelength[0].unit

            nonoverlapping_wavelengths.append(current_order.masked_wavelength[nonoverlapping].value)
            nonoverlapping_fluxes.append(current_order.masked_flux[nonoverlapping].value)

            current_overlapping = current_order.masked_wavelength > next_min
            next_overlapping = next_order.masked_wavelength < current_max

            # Does this order overlap with the next order?
            if np.count_nonzero(current_overlapping) > 0:
                # Find the overlapping parts between each order and the next order, and take
                # the mean of the two in the overlapping wavelength region, after interpolating
                # onto a common wavelength grid

                ol_wl_min = current_order.masked_wavelength[current_overlapping].min()
                ol_wl_max = current_order.masked_wavelength[current_overlapping].max()
                n_wl = 0.5 * (np.count_nonzero(current_overlapping) + np.count_nonzero(next_overlapping))

                common_wavelength_grid = np.linspace(ol_wl_min.value, ol_wl_max.value, int(n_wl))
                current_overlap_interp = np.interp(common_wavelength_grid,
                                                   current_order.masked_wavelength[current_overlapping].value,
                                                   current_order.masked_flux[current_overlapping].value)
                next_overlap_interp = np.interp(common_wavelength_grid,
                                                next_order.masked_wavelength[next_overlapping].value,
                                                next_order.masked_flux[next_overlapping].value)

                nonoverlapping_wavelengths.append(common_wavelength_grid)
                nonoverlapping_fluxes.append(np.mean([current_overlap_interp, next_overlap_interp], axis=0))

        nonoverlapping_fluxes = np.concatenate(nonoverlapping_fluxes)
        nonoverlapping_wavelengths = (np.concatenate(nonoverlapping_wavelengths)
                                      * dispersion_unit)

        return Spectrum1D(wavelength=nonoverlapping_wavelengths,
                          flux=nonoverlapping_fluxes, continuum_normalized=True,
                          mask=np.zeros_like(nonoverlapping_fluxes).astype(bool),
                          meta=dict(header=self.header))


def slice_spectrum(spectrum, min_wavelength, max_wavelength, norm=None):
    """
    Return a slice of a spectrum on a smaller wavelength range.

    Parameters
    ----------
    spectrum : `Spectrum1D`
        Spectrum to slice.
    min_wavelength : `~astropy.units.Quantity`
        Minimum wavelength to include in new slice
    max_wavelength : `~astropy.units.Quantity`
        Maximum wavelength to include in new slice
    norm : float
        Normalize the new slice fluxes by ``norm`` divided by the maximum flux
        of the new slice.

    Returns
    -------
    sliced_spectrum : `Spectrum1D`
    """
    in_range = ((spectrum.wavelength < max_wavelength) &
                (spectrum.wavelength > min_wavelength))

    wavelength = spectrum.wavelength[in_range]

    if norm is None:
        flux = spectrum.flux[in_range]
    else:
        flux = spectrum.flux[in_range] * norm / spectrum.flux[in_range].max()

    return Spectrum1D.from_array(wavelength, flux,
                                 dispersion_unit=spectrum.wavelength_unit)


def interpolate_spectrum(spectrum, new_wavelengths):
    """
    Linearly interpolate a spectrum onto a new wavelength grid.

    Parameters
    ----------
    spectrum : `Spectrum1D`
        Spectrum to interpolate onto new wavelengths
    new_wavelengths : `~astropy.units.Quantity`
        New wavelengths to interpolate the spectrum onto

    Returns
    -------
    interp_spec : `Spectrum1D`
        Interpolated spectrum.
    """

    sort_order = np.argsort(spectrum.masked_wavelength.to(u.Angstrom).value)
    sorted_spectrum_wavelengths = spectrum.masked_wavelength.to(u.Angstrom).value[sort_order]
    sorted_spectrum_fluxes = spectrum.masked_flux[sort_order]

    new_flux = np.interp(new_wavelengths.to(u.Angstrom).value,
                         sorted_spectrum_wavelengths,
                         sorted_spectrum_fluxes)

    return Spectrum1D.from_array(new_wavelengths, new_flux,
                                 dispersion_unit=spectrum.wavelength_unit)


def cross_corr(target_spectrum, model_spectrum, kernel_width):
    """
    Cross correlate an observed spectrum with a model.

    Convolve the model with a Gaussian kernel.

    Parameters
    ----------
    target_spectrum : `Spectrum1D`
        Observed spectrum of star
    model_spectrum : `Spectrum1D`
        Model spectrum of star
    kernel_width : float
        Smooth the model spectrum with a kernel of this width, in units of the
        wavelength step size in the model
    Returns
    -------
    wavelength_shift : `~astropy.units.Quantity`
        Wavelength shift required to shift the target spectrum to the rest-frame
    """

    smoothed_model_flux = gaussian_filter1d(model_spectrum.masked_flux,
                                            kernel_width)

    corr = np.correlate(target_spectrum.masked_flux - target_spectrum.masked_flux.mean(),
                        smoothed_model_flux - smoothed_model_flux.mean(), mode='same')

    max_corr_ind = np.argmax(corr)
    index_shift = corr.shape[0]/2 - max_corr_ind

    delta_wavelength = np.median(np.abs(np.diff(target_spectrum.masked_wavelength)))

    wavelength_shift = index_shift * delta_wavelength
    return wavelength_shift


def _poly_model(p, x):
    """
    Polynomial model for lstsq continuum normalization
    """
    x_mean = x.mean()
    return np.polyval(p, x - x_mean)


def _residuals(p, x, y):
    """
    Model residuals for lstsq continuum normalization
    """
    return _poly_model(p, x) - y



try:
    from .version import version as __version__
except ImportError:
    __version__ = ''

from .phoenix import *
from .spectra import *
from .utils import *
from .masking import *
from .activity import *
from .spectral_type import *




import numpy as np
from astroquery.vizier import Vizier
Vizier.ROW_LIMIT = 1e10   # Otherwise would only show first 50 values

__all__ = ['get_duncan_catalog', 'sindex_catalog', 'query_catalog_for_object',
           'get_k2_epic_catalog']

sindex_catalog = None
k2_epic_table = None
duncan1991 = 'III/159A'
huber2016 = 'J/ApJS/224/2/table5'


def get_duncan_catalog():
    """
    Parameters
    ----------

    Returns
    -------
    """
    global sindex_catalog

    if sindex_catalog is None:
        duncan1991 = 'III/159A'
        sindex_catalog = Vizier(catalog=duncan1991,
                                columns=["*", "Bmag", "Vmag"],
                                row_limit=1e10
                                ).query_constraints()[0]
    return sindex_catalog


def query_catalog_for_object(identifier, catalog=duncan1991):
    """
    Parameters
    ----------
    identifier : str

    catalog : str (optional)

    Returns
    -------

    """
    query = Vizier.query_object(identifier, catalog=catalog)

    if len(query) > 0:
        return query[0][0]
    else:
        return dict(Smean=np.nan, Smin=np.nan, Smax=np.nan)


def get_k2_epic_catalog():
    """
    Parameters
    ----------

    Returns
    -------
    """
    global k2_epic_table

    if k2_epic_table is None:
        catalogs = Vizier.get_catalogs(huber2016)
        k2_epic_table = catalogs[0]  # This is the table with the data
        k2_epic_table.add_index('EPIC')
    return k2_epic_table




import numpy as np
from astropy.utils.data import download_file
from astropy.io import fits
import astropy.units as u


__all__ = ['get_phoenix_model_spectrum', 'phoenix_model_temps']


phoenix_model_temps = np.array([3200, 6400, 14500, 4100, 12500, 5000, 6700,
                                 10400, 2700, 11600, 3600, 7000, 7600, 4500,
                                 10600, 5400, 5700, 3100, 7400, 10000, 4000,
                                 6000, 4900, 8400, 7200, 2600, 6300, 15000,
                                 3500, 4400, 6600, 9600, 13500, 5300, 7800,
                                 3000, 6900, 9400, 3900, 12000, 4800, 5600,
                                 2500, 9200, 8800, 3400, 5900, 4300, 8000,
                                 9000, 5200, 6200, 9800, 2900, 13000, 3800,
                                 6500, 11400, 14000, 4700, 2400, 6800, 8600,
                                 3300, 4200, 5500, 11800, 5100, 11000, 10200,
                                 2800, 5800, 3700, 8200, 10800, 4600, 6100,
                                 2300, 11200])


def get_url(T_eff, log_g):
    closest_grid_temperature = phoenix_model_temps[np.argmin(np.abs(phoenix_model_temps - T_eff))]

    url = ('ftp://phoenix.astro.physik.uni-goettingen.de/v2.0/HiResFITS/'
           'PHOENIX-ACES-AGSS-COND-2011/Z-0.0/lte{T_eff:05d}-{log_g:1.2f}-0.0.PHOENIX-'
           'ACES-AGSS-COND-2011-HiRes.fits').format(T_eff=closest_grid_temperature,
                                                    log_g=log_g)
    return url


def get_phoenix_model_spectrum(T_eff, log_g=4.5, cache=True):
    """
    Download a PHOENIX model atmosphere spectrum for a star with given
    properties.

    Parameters
    ----------
    T_eff : float
        Effective temperature. The nearest grid-temperature will be selected.
    log_g : float
        This must be a log g included in the grid for the effective temperature
        nearest ``T_eff``.
    cache : bool
        Cache the result to the local astropy cache. Default is `True`.

    Returns
    -------
    spectrum : `~specutils.Spectrum1D`
        Model spectrum
    """
    url = get_url(T_eff=T_eff, log_g=log_g)
    fluxes_path = download_file(url, cache=cache, timeout=30)
    fluxes = fits.getdata(fluxes_path)

    wavelength_url = ('ftp://phoenix.astro.physik.uni-goettingen.de/v2.0/HiResFITS/'
                      'WAVE_PHOENIX-ACES-AGSS-COND-2011.fits')
    wavelength_path = download_file(wavelength_url, cache=cache, timeout=30)
    wavelengths_vacuum = fits.getdata(wavelength_path)

    # Wavelengths are provided at vacuum wavelengths. For ground-based
    # observations convert this to wavelengths in air, as described in
    # Husser 2013, Eqns. 8-10:
    sigma_2 = (10**4 / wavelengths_vacuum)**2
    f = (1.0 + 0.05792105/(238.0185 - sigma_2) + 0.00167917 /
         (57.362 - sigma_2))
    wavelengths_air = wavelengths_vacuum / f

    mask_negative_wavelengths = wavelengths_air > 0

    from .spectra import Spectrum1D

    spectrum = Spectrum1D.from_array(wavelengths_air[mask_negative_wavelengths],
                                     fluxes[mask_negative_wavelengths],
                                     dispersion_unit=u.Angstrom)

    return spectrum




"""
Tools for measuring equivalent widths, S-indices.
"""

import numpy as np
import matplotlib.pyplot as plt

import astropy.units as u
from astropy.time import Time

from .catalog import query_catalog_for_object

__all__ = ['integrate_spectrum_trapz', 'true_h_centroid', 'true_k_centroid',
           'uncalibrated_s_index', 'StarProps', 'Measurement',
           'FitParameter']

true_h_centroid = 3968.4673 * u.Angstrom
true_k_centroid = 3933.6614 * u.Angstrom


def integrate_spectrum_trapz(spectrum, center_wavelength, width,
                             weighting=False, plot=False):
    """
    Integrate the area under a spectrum.

    Parameters
    ----------
    spectrum : `EchelleSpectrum`
        Spectrum to integrate under
    center_wavelength : `~astropy.units.Quantity`
        Center of region to integrate
    width : `~astropy.units.Quantity`
        Width about the center to integrate
    wavelength_offset : float
        Offset wavelengths by this amount, which is useful if a wavelength
        solution refinement as been made.
    weighting : bool
        Apply a triangular weighting function to the fluxes

    Returns
    -------
    integral : float
        Integral under the spectrum
    error : float
        Square-root of the sum of the fluxes within the bounds of the integral
    """
    wavelength = spectrum.wavelength
    wavelengths_increasing = wavelength[1] > wavelength[0]
    flux = spectrum.flux
    norm_const = spectrum.meta['normalization']

    if not wavelengths_increasing:
        wavelength = wavelength[::-1]
        flux = flux[::-1]

    # Assert fluxes are not negative
    flux.value[flux.value < 0] = 0.0

    if (not center_wavelength < wavelength.max() and
            not center_wavelength > wavelength.min()):
        raise ValueError("This spectral order does not contain"
                         "the center_wavelength given.")

    within_bounds = ((wavelength > center_wavelength - width/2) &
                     (wavelength < center_wavelength + width/2))

    if not weighting:
        integral = np.trapz(flux[within_bounds].value,
                            wavelength[within_bounds].value)

        lam = wavelength[within_bounds].value
        sigma_f = (np.sqrt(flux[within_bounds].value *
                           norm_const[within_bounds]) /
                   norm_const[within_bounds])
        sigma_f[np.isnan(sigma_f)] = np.nanmean(sigma_f)  # Fix negative fluxes
        a2_k = 0.25 * (lam[1:] - lam[:-1])**2 * (sigma_f[1:]**2 +
                                                 sigma_f[:-1]**2)
        error = np.sqrt(np.sum(a2_k))

    else:
        triangle_weights = triangle_weighting(wavelength[within_bounds],
                                              center_wavelength)
        integral = np.trapz(flux[within_bounds].value * triangle_weights,
                            wavelength[within_bounds].value)

        lam = wavelength[within_bounds].value
        sigma_f = (np.sqrt(flux[within_bounds].value *
                           norm_const[within_bounds]) * triangle_weights /
                   norm_const[within_bounds])
        sigma_f[np.isnan(sigma_f)] = np.nanmean(sigma_f)  # Fix negative fluxes
        a2_k = 0.25 * (lam[1:] - lam[:-1])**2 * (sigma_f[1:]**2 +
                                                 sigma_f[:-1]**2)
        error = np.sqrt(np.sum(a2_k))

    if plot:
        plt.figure()
        plt.plot(wavelength[spectrum.mask], flux[spectrum.mask])
        if weighting:
            triangle = triangle_weighting(wavelength[within_bounds],
                                          center_wavelength)
            plt.plot(wavelength[within_bounds], triangle, 'r', lw=2)
        plt.show()

    return Measurement(integral, err=error)


def triangle_weighting(x, x0, fwhm=1.09*u.Angstrom):
    """
    Compute the triangular weighting function used in CaII cores

    Parameters
    ----------
    x : `~astropy.units.Quantity`, array-like
        Wavelengths
    x0 : `~astropy.units.Quantity`
        Central wavelength of emission feature
    fwhm : `~astropy.units.Quantity`
        Full-width at half maximum of the weighting function
    """

    left_half = (x <= x0) & (x > x0 - fwhm)
    right_half = (x > x0) & (x < x0 + fwhm)

    weights = np.zeros_like(x.value)
    # float typecasting below resolves the unitless quantities
    weights[left_half] = ((x[left_half] - x0) / fwhm).value + 1
    weights[right_half] = ((x0 - x[right_half]) / fwhm).value + 1

    return weights


def uncalibrated_s_index(spectrum, plots=False):
    """
    Calculate the uncalibrated S-index from an Echelle spectrum.

    Parameters
    ----------
    spectrum : `EchelleSpectrum`
        Normalized target spectrum

    Returns
    -------
    s_ind : `SIndex`
        S-index. This value is intrinsic to the instrument you're using.
    """

    order_h = spectrum.get_order(89)
    order_k = spectrum.get_order(90)
    order_r = spectrum.get_order(91)
    order_v = spectrum.get_order(88)

    r_centroid = 3900 * u.Angstrom
    v_centroid = 4000 * u.Angstrom
    hk_fwhm = 1.09 * u.Angstrom
    hk_width = 2 * hk_fwhm
    rv_width = 20 * u.Angstrom

    h = integrate_spectrum_trapz(order_h, true_h_centroid, hk_width,
                                        weighting=True, plot=plots)
    k = integrate_spectrum_trapz(order_k, true_k_centroid, hk_width,
                                        weighting=True, plot=plots)
    r = integrate_spectrum_trapz(order_r, r_centroid, rv_width, plot=plots)
    v = integrate_spectrum_trapz(order_v, v_centroid, rv_width, plot=plots)

    s_ind = SIndex(h=h, k=k, r=r, v=v, time=spectrum.time)
    return s_ind


class SIndex(object):
    def __init__(self, h, k, r, v, k_factor=0.84, v_factor=1.0, time=None):
        """
        The pre-factors have been chosen to make the ``h`` and ``k`` values
        of the same order of magnitude; same for ``r`` and ``v``.

        Parameters
        -----------
        h : float
            CaII H feature emission flux
        k : float
            CaII K feature emission flux
        r : float
            Pseudo-continuum flux redward of CaII H&K
        v : float
            Pseudo-continuum flux blueward of CaII H&K
        k_factor : float
            Multiplicative factor for the K emission feature flux
            to make the H & K fluxes similar
        v_factor : float
            Multiplicative factor for the blue continuum region flux
            to make the r & v fluxes similar
        time : `~astropy.time.Time`
            Time this S-index measurement was taken.
        """
        self.r = r
        self.v = v
        self.h = h
        self.k = k

        self.k_factor = k_factor
        self.v_factor = v_factor

        self.time = time

    @property
    def uncalibrated(self):
        """
        Compute Eqn 2 of Isaacson+ 2010, for C1=1 and C2=0. This can be used
        to solve for C1 and C2.
        """
        uncalibrated_s_ind = ((self.h.value + self.k_factor * self.k.value) /
                              (self.r.value + self.v_factor * self.v.value))

        s_ind_err = (1 / (self.r.value + self.v_factor*self.v.value)**2 *
                     (self.h.err**2 + self.k_factor**2 * self.k.err**2) +
                     (self.h.value + self.k_factor*self.k.value)**2 /
                     (self.r.value + self.v_factor * self.v.value)**4 *
                     (self.r.err**2 + self.v_factor**2 * self.v.err**2)
                     )**0.5

        return Measurement(uncalibrated_s_ind, err=s_ind_err)

    def calibrated(self, c1, c2):
        """
        Calibrated S-index measurement (comparable with MWO S-indices).

        Uses the scaling constants as defined in Isaacson 2010+ (c1 and c2).

        Parameters
        ----------
        c1 : float
        c2 : float

        Returns
        -------
        Calibrated S-index.
        """
        return c1 * self.uncalibrated + c2

    @classmethod
    def from_dict(cls, dictionary):
        d = dictionary.copy()
        for key in dictionary:
            if key == 'time':
                d[key] = Time(float(d[key]), format='jd')
            elif key in ['h', 'k', 'r', 'v']:
                d[key] = Measurement.from_dict(d[key])
            else:
                d[key] = float(d[key])
        return cls(**d)

    def to_dict(self):
        d = dict()

        for attr in self.__dict__:
            value = getattr(self, attr)
            if isinstance(value, Measurement):
                value = value.__dict__
            elif isinstance(value, Time):
                value = str(value.jd)
            d[attr] = value
        return d


class StarProps(object):
    """
    S-index properties for a star
    """
    def __init__(self, name=None, s_apo=None, s_mwo=None, time=None):
        self.name = name
        self.s_apo = s_apo
        self._s_mwo = s_mwo
        self.time = time

    def get_s_mwo(self):
        obj = query_catalog_for_object(self.name)

        # Replace the uncertainty by twice the mean uncertainty
        # of the Duncan 1991 tables if no uncertainty is provided
        # (calculated by get_mean_mwo_error.py)
        error_Smean = obj['e_Smean'] if obj['e_Smean'] != 0 else 10 * 0.0205

        self._s_mwo = Measurement(obj['Smean'], err=error_Smean)

    @property
    def s_mwo(self):
        if self._s_mwo is None:
            self.get_s_mwo()
        return self._s_mwo

    @classmethod
    def from_dict(cls, dictionary):
        if dictionary['s_apo'] != 'None':
            if "value" in dictionary['s_apo']:
                s_apo = Measurement.from_dict(dictionary['s_apo'])
            else:
                s_apo = SIndex.from_dict(dictionary['s_apo'])
        else:
            s_apo = None

        if '_s_mwo' in dictionary and dictionary['_s_mwo'] != "None":
            s_mwo = Measurement.from_dict(dictionary['_s_mwo'])
        else:
            s_mwo = None

        if dictionary['time'] != 'None':
            dictionary['time'] = Time(float(dictionary['time']), format='jd')
        else:
            dictionary['time'] = None

        return cls(s_apo=s_apo, s_mwo=s_mwo, name=dictionary['name'],
                   time=dictionary['time'])


class Measurement(object):
    def __init__(self, value=None, err=None, time=None, default_err=1e10,
                 meta=None):

        if hasattr(value, '__len__'):
            self.value = np.asarray(value)
            self.err = np.asarray(err)

            self.err[self.err == 0] = default_err

        else:
            self.value = value
            if err == 0:
                self.err = default_err
            else:
                self.err = err

        if isinstance(time, Time):
            self.time = time
        elif hasattr(time, 'real'):  # if time is float
            self.time = Time(time, format='jd')
        elif isinstance(time, list):
            if hasattr(time[0], 'real'):
                self.time = Time(time, format='jd')
            else:
                self.time = Time(time)
        else:
            self.time = None

        self.meta = meta

    @classmethod
    def from_min_max(cls, min, max):
        mean = 0.5*(min + max)
        return cls(value=mean, err=max-mean)

    @classmethod
    def from_dict(cls, dictionary):
        kwargs = {key: float(dictionary[key])
                  if (dictionary[key] != "None" and
                      dictionary[key] is not None)
                  else None
                  for key in dictionary}
        return cls(**kwargs)

    def __repr__(self):
        return "<{0}: {1} +/- {2}>".format(self.__class__.__name__,
                                          self.value, self.err)

    def __getitem__(self, item):

        new_attrs = dict()

        for attr in ['value', 'err', 'time']:
            attr_value = getattr(self, attr)
            if attr_value is not None and hasattr(attr_value, '__len__'):
                new_attrs[attr] = attr_value[item]

        return Measurement(**new_attrs)

    def __len__(self):
        return len(self.value)

    def to_latex(self):
        return "${0:.3f} \pm {1}$".format(self.value, error_to_latex(self.err))


def error_to_latex(error):
    str_err = "{0:.2g}".format(error)
    if 'e' in str_err:
        str_err = "{0:.6f}".format(error)
    return str_err


class FitParameter(object):
    """
    Fit results for a fitting parameter, with asymmetrical errors.
    """
    def __init__(self, value, err_upper=None, err_lower=None, default_err=1e10):
        """

        Parameters
        ----------
        value : {`~astropy.units.Quantity`, `~numpy.ndarray`, float}
        err_upper : {`~astropy.units.Quantity`, `~numpy.ndarray`, float}
        err_lower : {`~astropy.units.Quantity`, `~numpy.ndarray`, float}
        default_err : float
        """
        if hasattr(value, '__len__'):
            value = np.asarray(value)
            err_upper = np.asarray(err_upper)
            err_lower = np.asarray(err_lower)

            self.value = value
            self.err_upper = err_upper
            self.err_lower = err_lower

        else:

            self.value = value
            if err_upper == 0 or err_lower == 0:
                self.err_upper = default_err
                self.err_lower = default_err
            else:
                self.err_upper = err_upper
                self.err_lower = err_lower

    @classmethod
    def from_text(cls, path):
        value, err_upper, err_lower = np.loadtxt(path, unpack=True)
        return cls(value, err_lower=err_lower, err_upper=err_upper)

    def to_text(self, path):
        np.savetxt(path, [self.value, self.err_upper, self.err_lower])

    def __repr__(self):
        return "<{0}: {1} +{2} -{3}>".format(self.__class__.__name__,
                                             self.value, self.err_upper,
                                             self.err_lower)



import numpy as np
from scipy.optimize import fmin_l_bfgs_b
import matplotlib.pyplot as plt

__all__ = ['get_spectrum_mask']

def _gaussian(x, a, x0, sigma):
    return a * np.exp(-0.5 * (x - x0)**2 / sigma**2)


def _chi2(p, x, y):
    return np.sum((_gaussian(x, *p) - y) ** 2)


def get_spectrum_mask(spectrum, cutoff=1.5, plot=False):
    """
    Fit the raw, pre-normalized spectrum's blaze function
    with a Gaussian. Use a

    Parameters
    ----------
    spectrum : `~specutils.Spectrum1D`
        Spectrum to fit
    cutoff : float
        Mask channels greater than ``cuttoff``-sigma away from
        the peak flux.

    Returns
    -------
    mask : `~numpy.ndarray`
        Mask that excludes channels greater than ``cuttoff``-sigma away from
        the peak flux.
    """
    wavelength_ptp = np.ptp(spectrum.wavelength.value)
    initp = np.array([spectrum.flux.max().value,
                      spectrum.wavelength.mean().value,
                      wavelength_ptp/4])

    bestp = fmin_l_bfgs_b(_chi2, initp[:],
                          approx_grad=True,
                          args=(spectrum.wavelength.value,
                                spectrum.flux.value),
                          bounds=[(0, np.inf),
                                  (spectrum.wavelength.min().value,
                                   spectrum.wavelength.max().value),
                                  (wavelength_ptp/8,
                                   wavelength_ptp/2)])[0]

    best_a, best_x0, best_sigma = bestp

    mask = ((spectrum.wavelength.value > best_x0 - cutoff * best_sigma) &
            (spectrum.wavelength.value < best_x0 + cutoff * best_sigma))

    if plot:
        plt.figure()
        plt.plot(spectrum.wavelength, spectrum.flux, label='unmasked')
        plt.plot(spectrum.wavelength[mask], spectrum.flux[mask], label='masked')
        plt.legend()

    return np.logical_not(mask)



#!/usr/bin/env python
import os
from setuptools import setup

VERSION_TEMPLATE = """
# Note that we need to fall back to the hard-coded version if either
# setuptools_scm can't be imported or setuptools_scm can't determine the
# version, so we catch the generic 'Exception'.
try:
    from setuptools_scm import get_version
    version = get_version(root='..', relative_to=__file__)
except Exception:
    version = '{version}'
""".lstrip()

setup(use_scm_version={'write_to': os.path.join('aesop', 'version.py'),
                       'write_to_template': VERSION_TEMPLATE})
